def test_tutorial001(add_docs_to_pythonpath):
    from docs_src.GettingStarted.tutorial001 import run as run001

    _ = run001()


def test_tutorial002(add_docs_to_pythonpath):
    from docs_src.GettingStarted.tutorial001 import run as run002

    _ = run002()
