def test_tutorial002(add_docs_to_pythonpath):
    from docs_src.MelusineDetectors.tutorial001 import run as run001

    _ = run001()


def test_tutorial002(add_docs_to_pythonpath):
    from docs_src.MelusineDetectors.tutorial002 import run as run002

    _ = run002()


def test_tutorial003(add_docs_to_pythonpath):
    from docs_src.MelusineDetectors.tutorial003 import run as run003

    _ = run003()


def test_tutorial004(add_docs_to_pythonpath):
    from docs_src.MelusineDetectors.tutorial004 import run as run004

    _ = run004()
